--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- battle.lua: Battle mod abilities
--

local battleLoaded = false

freeslot(
	"SPR2_SURS",
	"SPR2_SURG",

	"S_PLAY_PLAGUESURGESTART",
	"S_PLAY_PLAGUESURGELOOP",

	"MT_PLAGUE_SURGEHITBOX",

	"S_PLAGUE_SURGEPROJ",
	"SPR_SURP",

	"sfx_pksurg"
)

spr2defaults[SPR2_SURS] = SPR2_JUMP
spr2defaults[SPR2_SURG] = SPR2_JUMP

states[S_PLAY_PLAGUESURGESTART] = {SPR_PLAY, SPR2_SURS|FF_SPR2ENDSTATE, 2, nil, S_PLAY_PLAGUESURGELOOP, 0, S_PLAY_PLAGUESURGESTART}
states[S_PLAY_PLAGUESURGELOOP] = {SPR_PLAY, SPR2_SURG, 2, nil, 0, 0, S_PLAY_PLAGUESURGELOOP}

mobjinfo[MT_PLAGUE_SURGEHITBOX] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 56*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOSECTOR|MF_SOLID|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

states[S_PLAGUE_SURGEPROJ] = {SPR_SURP, FF_FULLBRIGHT|FF_ANIMATE, 7*3, nil, 7, 3, S_NULL}

mobjinfo[MT_PLAGUE_BOMB].name = "plague bomb"
mobjinfo[MT_PLAGUE_EXPLOSION].name = "plague explosion"
mobjinfo[MT_PLAGUE_SURGEHITBOX].name = "staff of surging"

sfxinfo[sfx_pksurg].caption = "Staff of Surging"

local function surgeCollide(hitbox, collided)
	if not (CBW_Battle)
		return false
	end

	local owner = hitbox.target
	if not (owner and owner.valid)
		return false
	end

	if (collided == owner or collided.type == MT_PLAGUE_SURGEHITBOX)
		return false
	end

	if (hitbox.z + hitbox.height < collided.z)
	or (hitbox.z > collided.z + collided.height)
		return false
	end

	if (collided.battleobject) or (collided.player)
		-- We are simply extending the player interaction hitbox.
		collided.pushed = owner
		owner.pushed = collided
		--CBW_Battle.DoPlayerInteract(collided, owner)
	elseif (collided.flags & MF_SHOOTABLE)
		-- Generic damage
		P_DamageMobj(collided, hitbox, owner, 1, 0)
	end

	return false
end

addHook("MobjCollide", surgeCollide, MT_PLAGUE_SURGEHITBOX)
addHook("MobjMoveCollide", surgeCollide, MT_PLAGUE_SURGEHITBOX)

addHook("ShouldDamage", function(mo, inf, src)
	if (inf and inf.valid) and (inf.type == MT_PLAGUE_SURGEHITBOX)
		if not (CBW_Battle.PlayerCanBeDamaged(mo.player))
			return false
		end

		if (src and src.valid)
			if (mo == src)
				return false
			end

			if (src.player and src.player.valid)
			and (CBW_Battle.MyTeam(src.player, mo.player))
				return false
			end
		end
	end
end, MT_PLAYER)

addHook("MobjDamage", function(mo, inf, src)
	if not (CBW_Battle)
		return
	end

	if (inf and inf.valid) and (inf.type == MT_PLAGUE_SURGEHITBOX)
		local owner = inf.target
		CBW_Battle.PlayerCreditPusher(mo.player, owner)
	end
end, MT_PLAYER)

addHook("TouchSpecial", function(ring, mo)
	if (mo.skin == "plagueknight" and mo.actionstate == 1)
		-- Prevent stealing every single one of the other player's rings by using this move once
		return true
	end
end, MT_FLINGRING)

local function initPlagueBattle()
	if (battleLoaded == true)
		return
	end

	if not (CBW_Battle)
		return
	end

	battleLoaded = true

	local ACTION_NONE = 0
	local ACTION_UPPERCUT = 1
	local ACTION_COOLDOWN = 2

	local DOACTION_UNABLE = -1
	local DOACTION_ABLE = 0
	local DOACTION_PRESSED = 1
	local DOACTION_HELD = 2

	local plagueSpecials = function(mo, doaction)
		local player = mo.player

		player.actiontext = "Staff of Surging"
		player.actionrings = 10

		if not (CBW_Battle.CanDoAction(player))
			player.actionstate = ACTION_NONE
			return
		end

		if (player.actionstate == ACTION_NONE)
			if (doaction == DOACTION_PRESSED)
			and (mo.plague != nil)
			and (mo.plague.bombthrowing == 0)
				player.actionstate = ACTION_UPPERCUT
				player.actiontime = 0

				CBW_Battle.PayRings(player)
				S_StartSound(mo, sfx_pksurg)

				mo.plague.surgeangle = mo.angle
			end
		end

		if (player.actionstate == ACTION_UPPERCUT)
			mo.plague.bombburst = false

			player.skidtime = 0
			mo.plague.skidtime = 0

			player.drawangle = mo.plague.surgeangle

			if (mo.state == S_PLAY_PLAGUESURGELOOP)
				player.actiontime = $1+1

				if (player.actiontime >= 10)
					mo.state = S_PLAY_FALL
					P_SetObjectMomZ(mo, 8*FRACUNIT, false)
					player.actionstate = ACTION_COOLDOWN
					player.actiontime = 0
				else
					local momz = 6*FRACUNIT

					if (mo.eflags & MFE_UNDERWATER)
						momz = FixedDiv($1, FixedSqrt(3*FRACUNIT))
					end

					momz = $1 * (10 - player.actiontime)

					P_SetObjectMomZ(mo, momz, false)
					P_InstaThrust(mo, mo.plague.surgeangle, 4*mo.scale)

					local hoffset = FixedDiv(mo.height, mo.scale) - (24*FRACUNIT)
					local hitbox = P_SpawnMobjFromMobj(
						mo,
						0,
						0,
						hoffset,
						MT_PLAGUE_SURGEHITBOX
					)
					hitbox.target = mo
					hitbox.tics = 6*3

					hitbox.surgex = mo.momx
					hitbox.surgey = mo.momy
					hitbox.surgez = mo.momz

					local offset = (ANGLE_180 / 6)
					local dist = 3 * FixedDiv(mo.radius, mo.scale)

					for i = 1,2
						local anglechange = offset * (((10 - player.actiontime) * 2) + i)
						local angle = mo.plague.surgeangle + anglechange

						local particle = P_SpawnMobjFromMobj(
							mo,
							FixedMul(dist, cos(angle)),
							FixedMul(dist, sin(angle)),
							hoffset,
							MT_PLAGUE_SURGEAURA
						)

						particle.state = S_PLAGUE_SURGEPROJ
						particle.tics = $1 - (2 - i)
						particle.color = skincolors[player.skincolor].invcolor
					end
				end
			else
				if (mo.state != S_PLAY_PLAGUESURGESTART)
					mo.state = S_PLAY_PLAGUESURGESTART
				end
			end
		end

		if (player.actionstate == ACTION_COOLDOWN)
			-- Cooldown state, only reverts when hitting the ground.
			if (P_IsObjectOnGround(mo))
				player.actionstate = ACTION_NONE
				player.actiontime = 0
			end
		end

		if (player.actionstate != ACTION_NONE)
			-- Spawn aura during usage / cooldown
			local radius = (mo.radius / mo.scale) * 2
			local aura = P_SpawnMobjFromMobj(
				mo,
				P_RandomRange(-radius, radius) * FRACUNIT,
				P_RandomRange(-radius, radius) * FRACUNIT,
				(P_RandomRange(-radius, radius) * FRACUNIT) + FixedDiv(mo.height / 2, mo.scale),
				MT_PLAGUE_SURGEAURA
			)

			aura.color = skincolors[player.skincolor].invcolor
		end
	end

	local plaguePriority = function(player)
		if (player.mo and player.mo.valid and player.mo.plague != nil)
			if (player.actionstate == ACTION_UPPERCUT)
				CBW_Battle.SetPriority(
					player,
					3, 3,
					nil, 3, 3,
					mobjinfo[MT_PLAGUE_SURGEHITBOX].name
				)
			elseif (player.mo.plague.bombburst == true)
				CBW_Battle.SetPriority(
					player,
					2, 1,
					nil, 2, 1,
					"bomb burst"
				)
			end
		end
	end

	local plagueCollide = function(
		n1, n2,
		plr, mo,
		atk, def, weight,
		hurt, pain, ground,
		angle, thrust, thrust2, collisiontype
	)
		if (plr[n1] and plr[n1].valid and mo[n1].plague != nil)
			if (plr[n1].actionstate == ACTION_UPPERCUT and hurt == 1)
				-- Successful surge attack.
				-- Knock them into the air.
				P_InstaThrust(mo[n2], angle[n2], mo[n1].scale * 6)
				CBW_Battle.ZLaunch(mo[n2], mo[n1].scale * 24, false)
				return true
			end
		end

		return false
	end

	CBW_Battle.SkinVars["plagueknight"] = {
		flags = SKINVARS_GUARD|SKINVARS_NOSPINSHIELD,
		weight = 80,
		guard_frame = 1,
		special = plagueSpecials,
		func_priority_ext = plaguePriority,
		func_collide = plagueCollide
	}
end

--[[
initPlagueBattle()

addHook("MapLoad", function()
	-- Doing this on MapLoad so that
	-- we don't have a ThinkFrame running all of the time :)
	initPlagueBattle()
end)
--]]

addHook("ThinkFrame", function()
	-- damn
	initPlagueBattle()
end)
