--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- bombs.lua: Bomb throwing ability & projectiles
--

Plague.ThrowBomb = function(player, onground)
	if not (Plague.AllowAbilities(player, PF_STASIS, false))
		return
	end

	for i = 1,Plague.MaxBombs(player)
		if (player.mo.plague.bombs[i] and player.mo.plague.bombs[i].valid)
			continue
		end

		Plague.Reset(player);

		-- Slow momentum, restore control
		player.mo.momx = 2 * $1 / 3
		player.mo.momy = 2 * $1 / 3

		if not (onground)
			local hop = Plague.ScaleMomZ(player.mo, 5*FRACUNIT)
			if (player.mo.momz * P_MobjFlip(player.mo) < hop)
				P_SetObjectMomZ(player.mo, hop, false)
			end
		end

		if (onground)
			player.mo.state = S_PLAY_PLAGUETHROWGROUND
		else
			player.mo.state = S_PLAY_PLAGUETHROWAIR
		end

		player.mo.plague.throwangle = player.mo.angle

		player.mo.plague.bombthrowing = 5
		if (player.powers[pw_super])
			player.mo.plague.bombthrowcool = 5
		else
			player.mo.plague.bombthrowcool = 10
		end
		player.mo.plague.bombburst = false

		S_StartSound(player.mo, sfx_pkthro)
		break
	end
end

Plague.BombAim = function(player, onground)
	local defaultHorzAngle = player.mo.plague.throwangle
	local defaultVertAngle = 0
	local maxSearchCone = ANG10
	local twod = ((maptol & TOL_2D) or (player.mo.flags2 & MF2_TWOD) or (player.mo.currentaxis != nil))

	if (onground == false)
		if (twod)
			defaultVertAngle = -ANGLE_45
		else
			defaultVertAngle = -ANGLE_22h
		end

		defaultVertAngle = $1 * P_MobjFlip(player.mo)
		maxSearchCone = $1 * 2
	end

	local finalAngles = {
		ha = defaultHorzAngle,
		va = defaultVertAngle
	}

	if (twod)
		-- No need for the autoaim in a 2D space.
		return finalAngles;
	end

	local blockmapSearchDist = 512 * player.mo.scale
	local maxDistance = blockmapSearchDist*2
	local center = {
		x = player.mo.x + FixedMul(blockmapSearchDist, cos(defaultHorzAngle)),
		y = player.mo.y + FixedMul(blockmapSearchDist, sin(defaultHorzAngle)),
	}

	local highScore = 0

	searchBlockmap("objects", function(ref, found)
		-- Hittable currently
		if (found.health <= 0) or (found.flags & MF_NOCLIPTHING) or (found.flags2 & MF2_FRET)
			return
		end

		-- Check for aimable
		local aimable = false

		if ((found.flags & MF_SHOOTABLE) and (found.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)))
		or ((found.player and found.player.valid))
			aimable = true
		end

		if (found.flags2 & MF2_INVERTAIMABLE)
			aimable = (not aimable)
		end

		if (aimable == false)
			return
		end

		-- too far away?
		local fflip = P_MobjFlip(found)
		local fhalf = found.height / 2
		local fz = found.z + fhalf

		if (onground == true)
			-- Aim for their top, when
			-- throwing bombs from the ground.
			fz = $1 + (fflip * fhalf)
		else
			-- Aim for their bottom, when
			-- throwing bombs from the air.
			fz = $1 - (fflip * fhalf)
		end

		local xydist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
		local zdist = fz - (ref.z + (ref.height/2))
		local fullDistance = FixedHypot(xydist, zdist)

		if (fullDistance > maxDistance)
			return
		end

		local distMul = FixedDiv(maxDistance - fullDistance, maxDistance)
		local searchCone = maxSearchCone --FixedAngle(FixedMul(AngleFixed(maxSearchCone), distMul))

		-- Now we can compare angles
		local potentialAngles = {
			ha = R_PointToAngle2(ref.x, ref.y, found.x, found.y),
			va = R_PointToAngle2(0, 0, xydist, zdist)
		}

		local hAngleDiff = abs(defaultHorzAngle - potentialAngles.ha)
		local vAngleDiff = abs(defaultVertAngle - potentialAngles.va)

		if (hAngleDiff > searchCone or vAngleDiff > searchCone)
			-- Out of auto aim cone
			return
		end

		local score = 0

		if (gametyperules & GTR_FRIENDLY)
			-- Cooperative
			-- Prioritize enemies / bosses, then players, then monitors
			if (found.flags & (MF_ENEMY|MF_BOSS|MF_PAIN))
				score = $1 + FRACUNIT
			elseif (found.player and found.player.valid)
				score = $1 + FRACUNIT / 2
			end
		else
			-- Ringslinger / competitive
			-- Prioritize players, then enemies / bosses, then monitors
			if (found.player and found.player.valid)
				score = $1 + FRACUNIT
			elseif (found.flags & (MF_ENEMY|MF_BOSS|MF_PAIN))
				score = $1 + FRACUNIT / 2
			end
		end

		-- Closer to the origin = better
		score = $1 + distMul

		-- Closer to original angle = better
		score = $1 + FixedDiv(searchCone - hAngleDiff, searchCone) + FixedDiv(searchCone - vAngleDiff, searchCone)

		--[[
		print(string.format("ha: %f", AngleFixed(potentialAngles.ha)))
		print(string.format("va: %f", AngleFixed(potentialAngles.va)))
		print(string.format("hadiff: %f", AngleFixed(hAngleDiff)))
		print(string.format("vadiff: %f", AngleFixed(vAngleDiff)))
		print(string.format("score: %f", score))
		print("========")
		--]]

		if (score > highScore)
			finalAngles = potentialAngles
			highScore = score
		end
	end, player.mo, center.x - blockmapSearchDist, center.x + blockmapSearchDist, center.y - blockmapSearchDist, center.y + blockmapSearchDist)

	return finalAngles
end

Plague.SpawnBomb = function(player, onground)
	for i = 1,Plague.MaxBombs(player)
		if (player.mo.plague.bombs[i] and player.mo.plague.bombs[i].valid)
			continue
		end

		player.mo.plague.bombs[i] = P_SpawnMobjFromMobj(player.mo, 0, 0, FixedDiv(player.mo.height/2, player.mo.scale), MT_PLAGUE_BOMB)
		player.mo.plague.bombs[i].shadowscale = 2*FRACUNIT/3
		player.mo.plague.bombs[i].target = player.mo
		player.mo.plague.bombs[i].fuse = player.mo.plague.bombs[i].info.damage

		if (player and player.valid)
			player.mo.plague.bombs[i].color = player.skincolor
		else
			player.mo.plague.bombs[i].color = SKINCOLOR_SEAFOAM
		end

		local speed = FixedMul(player.mo.plague.bombs[i].info.speed, player.mo.scale) * 4

		if (onground == false)
			speed = $1 * 2
		end

		if (maptol & TOL_2D)
			speed = $1 / 2
		end

		local angles = Plague.BombAim(player, onground)

		player.mo.plague.bombs[i].angle = angles.ha

		local newMom = {
			x = FixedMul(FixedMul(speed, cos(angles.ha)), cos(angles.va)),
			y = FixedMul(FixedMul(speed, sin(angles.ha)), cos(angles.va)),
			z = FixedMul(speed, sin(angles.va)),
		}

		newMom.z = Plague.ScaleMomZ(player.mo, $1)

		player.mo.plague.bombs[i].momx = newMom.x + player.mo.momx
		player.mo.plague.bombs[i].momy = newMom.y + player.mo.momy
		player.mo.plague.bombs[i].momz = newMom.z + player.mo.momz

		/*
		if (Plague.ExplodeSearch(player.mo.plague.bombs[i], false) == true)
			P_KillMobj(player.mo.plague.bombs[i]) -- Gone as soon as it got here...
		end
		*/

		break
	end
end

addHook("MobjThinker", function(mo)
	if not (mo.health)
		return
	end

	if (Plague.ExplodeSearch(mo, true) == true)
		P_KillMobj(mo)
		return
	end

	mo.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)

	if (P_IsObjectOnGround(mo))
		P_InstaThrust(mo, mo.angle, FixedMul(mo.info.speed, mo.scale))
		Plague.ApplyMomZ(mo, 4*FRACUNIT)
	end
end, MT_PLAGUE_BOMB)

addHook("MobjMoveBlocked", function(mo)
	if (Plague.ExplodeSearch(mo, true) == true)
		P_KillMobj(mo)
	end
end, MT_PLAGUE_BOMB)

addHook("MobjFuse", function(mo)
	if (mo.health)
		P_KillMobj(mo)
	end
end, MT_PLAGUE_BOMB)

local function bombCollide(bomb, mo)
	if (bomb.health <= 0)
		return false
	end

	if (mo.z + mo.height < bomb.z)
	or (mo.z > bomb.z + bomb.height)
		return false
	end

	if (Plague.ExplosionDamage(bomb, mo))
		P_KillMobj(bomb)
	end

	return false
end

addHook("MobjCollide", bombCollide, MT_PLAGUE_BOMB)
addHook("MobjMoveCollide", bombCollide, MT_PLAGUE_BOMB)

addHook("MobjDeath", function(mo)
	Plague.SpawnPlagueExplosion(mo, mo.target, 0, 0, FixedDiv(mo.height/2, mo.scale), 0, false)
end, MT_PLAGUE_BOMB)
