--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- burst.lua: Bomb burst ability
--

Plague.BombBurst = function(player)
	if (player.mo.plague.spinbt)
		if (player.mo.plague.burstcharge < Plague.BURSTTHRESHOLD)
			player.mo.plague.burstcharge = $1+1
		end
	else
		if (player.mo.plague.burstcharge >= Plague.BURSTTHRESHOLD)
		and (Plague.AllowAbilities(player, PF_JUMPSTASIS, true))
			Plague.Reset(player);

			-- Don't allow jump momentum cutting
			player.pflags = $1 & ~(PF_JUMPED|PF_STARTJUMP)
			player.mo.z = $1 + P_MobjFlip(player.mo)

			local maxhthrust = 52 * player.mo.scale
			local minvthrust = 14 * FRACUNIT
			local addvthrust = 2 * FRACUNIT

			local inputdirection = 0

			local horzcomponent = 0
			local horizontalthrust = 0
			local verticalthrust = 0

			local simplecontrols = false

			if (player.pflags & PF_ANALOGMODE)
				simplecontrols = true
			end

			if (player.mo.flags2 & MF2_TWOD) or (twodlevel)
			or (player.mo.currentaxis)
				local s = player.cmd.sidemove

				if (player.mo.currentaxis)
					-- Axis2D support
					if (player.mo.controlflip)
						s = -s;
					end

					local axisAngle = 0;
					if (player.mo.currentaxis.angle != nil)
						axisAngle = player.mo.currentaxis.angle - ANGLE_90;
					else
						axisAngle = R_PointToAngle2(
							player.mo.currentaxis.x, player.mo.currentaxis.y,
							player.mo.x, player.mo.y
						);
					end

					inputdirection = axisAngle + R_PointToAngle2(0, 0, 0, s * FRACUNIT);

					if (player.mo.currentaxis.flipped)
						inputdirection = $1 + ANGLE_180;
					end
				else
					inputdirection = R_PointToAngle2(0, 0, s * FRACUNIT, 0)
				end

				minvthrust = 11 * $1 / 10
				addvthrust = 11 * $1 / 10

				if (maptol & TOL_2D)
					-- For SOME reason, this isn't done for mid-level 2D mode switch, only for explicit 2D maps.
					-- (I kinda wish it was always done this way but that's not my decision)
					maxhthrust = 2 * $1 / 3
				end

				horzcomponent = abs(player.cmd.sidemove)
				horizontalthrust = (maxhthrust / MAXPLMOVE) * horzcomponent
				verticalthrust = minvthrust + ((addvthrust / MAXPLMOVE) * (MAXPLMOVE - horzcomponent))
			else
				if (simplecontrols == true)
					-- Simple controls makes it work a little bit more like 2D mode.
					-- I think this is kind of bad but maybe some people will prefer this...?

					inputdirection = player.mo.angle
					maxhthrust = 2 * $1 / 3

					horzcomponent = FixedHypot(player.cmd.forwardmove, player.cmd.sidemove)
					horizontalthrust = (maxhthrust / MAXPLMOVE) * horzcomponent
					verticalthrust = minvthrust + ((addvthrust / MAXPLMOVE) * (MAXPLMOVE - horzcomponent))
				else
					inputdirection = player.mo.angle + R_PointToAngle2(
						0,
						0,
						player.cmd.forwardmove * FRACUNIT,
						-player.cmd.sidemove * FRACUNIT
					)

					horzcomponent = (player.cmd.forwardmove + MAXPLMOVE)
					horizontalthrust = (maxhthrust / (MAXPLMOVE*2)) * horzcomponent
					verticalthrust = minvthrust + ((addvthrust / (MAXPLMOVE*2)) * ((MAXPLMOVE*2) - horzcomponent))
				end
			end

			if (player.powers[pw_super])
				horizontalthrust = 3 * $1 / 2
				verticalthrust = 3 * $1 / 2
			end

			if (player.pflags & PF_STASIS)
				-- Allow jumping without PF_JUMPSTASIS, but you don't go forward.
				horizontalthrust = 0
			end

			if (Plague.DoesntHaveFlag(player) == false)
				-- Flag carriers are not meant to have ANY mobility options,
				-- but it'd be too weird to remove Plague Knight's main thing.

				-- Instead, we will severely gimp it, so the main utility is
				-- the explosion, not the jump.

				horizontalthrust = $1 / 3
				verticalthrust = 2 * $1 / 3
			end

			if (horizontalthrust == 0)
				player.mo.momx = 0
				player.mo.momy = 0
			else
				P_InstaThrust(player.mo, inputdirection, horizontalthrust)
			end

			if (verticalthrust > 0)
				Plague.ApplyMomZ(player.mo, verticalthrust)
			end

			player.mo.state = S_PLAY_PLAGUEBURST
			player.mo.plague.bombburst = true
			player.mo.plague.bombthrowcool = TICRATE/2
			player.mo.plague.burstinvuln = Plague.BURSTINVULTIME

			-- Now, for all of the explosion particles...
			Plague.SpawnPlagueExplosion(player.mo, player.mo, 0, 0, 192*FRACUNIT, 0, true)
			Plague.SpawnPlagueExplosion(player.mo, player.mo, 0, 0, 96*FRACUNIT, 1, true)
			Plague.SpawnPlagueExplosion(player.mo, player.mo, 0, 0, 0, 3, true)
			for i = 0,3
				local angle = player.mo.angle + ANGLE_45 + (ANGLE_90*i)
				Plague.SpawnPlagueExplosion(player.mo, player.mo,
					112*cos(angle), 112*sin(angle),
					48*FRACUNIT, 5, true
				)
			end

			S_StartSound(player.mo, sfx_pkburs)
		end

		player.mo.plague.burstcharge = 0
	end
end
