--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- explosion.lua: Bomb & burst explosion interactions
--

Plague.BustFOFsInSector = function(explode, sector)
	if (sector == nil)
		return false;
	end

	local foundone = false;

	for rover in sector.ffloors()
		if not ((rover.flags & FF_EXISTS) and (rover.flags & FF_BUSTUP))
			continue;
		end

		local b = rover.bottomheight;
		local t = rover.topheight;

		if (rover.t_slope and rover.t_slope.valid)
			t = P_GetZAt(rover.t_slope, explode.x, explode.y);
		end

		if (rover.b_slope and rover.b_slope.valid)
			b = P_GetZAt(rover.b_slope, explode.x, explode.y);
		end

		if (t < explode.z + explode.momz)
		or (b > explode.z + explode.height + explode.momz)
			continue;
		end

		EV_CrumbleChain(nil, rover);

		if (rover.master.flags & ML_EFFECT5)
			P_LinedefExecute(
				P_AproxDistance(rover.master.dx, rover.master.dy) >> FRACBITS,
				(explode.target and explode.target.valid) and explode.target or explode,
				sector
			);
		end

		foundone = true;
	end

	return foundone;
end

Plague.CoopBombBurst = function(mo, explode)
	local player = mo.player

	if not (player and player.valid)
		return false
	end

	if (player.pflags & PF_JUMPSTASIS)
	or (player.powers[pw_carry] != CR_NONE)
	or (mo.flags & MF_NOCLIP) or (mo.flags & MF_NOCLIPTHING)
		return false
	end

	if (mo.coopburst == nil)
		mo.coopburst = {}
	else
		if (mo.coopburst.time > 0)
			return false
		end
	end

	-- Give friendly players a jump boost!
	player.pflags = ($1|PF_JUMPED) & ~PF_STARTJUMP

	local power = 16*FRACUNIT
	if ((mo.flags2 & MF2_TWOD) or (twodlevel) or (mo.currentaxis))
		power = 11 * $1 / 10
	end

	Plague.ApplyMomZ(mo, power)

	if (player.pflags & PF_SPINNING)
		-- give them max control rolling!
		player.pflags = $1|PF_THOKKED
	else
		player.pflags = $1|P_GetJumpFlags(player)
		mo.state = S_PLAY_JUMP

		if (mo.skin != "plagueknight")
			-- give them their abilities back!
			player.pflags = $1 & ~(PF_THOKKED|PF_SHIELDABILITY)
		end
	end

	mo.coopburst.time = 7*Plague.BURSTTHRESHOLD/4 -- x1.75
	mo.coopburst.color = explode.color

	S_StartSound(mo, sfx_pkburs, player)

	return true
end

Plague.ExplosionDamage = function(explode, mo)
	local owner = explode.target

	if (owner == mo) or (explode == mo)
		-- Obviously don't damage yourself.
		return false
	end

	if (mo.type == MT_PLAYER)
		if (Plague.PlayerIsFriendly(owner.player, mo.player) == true)
			return Plague.CoopBombBurst(mo, explode)
		end

		if (mo.player.powers[pw_flashing])
			return false
		end

		if (CBW_Battle)
			if not (CBW_Battle.PlayerCanBeDamaged(mo.player))
				return false
			end
		end

		P_DamageMobj(mo, explode, owner, 1)
		return true
	elseif (mo.type == MT_EGGSHIELD)
		local delta = abs(R_PointToAngle2(mo.x, mo.y, explode.x - explode.momx, explode.y - explode.momy) - mo.angle)

		if (delta > ANGLE_90)
			-- damage the shield
			S_StartSound(explode, mo.info.deathsound)
			P_KillMobj(mo, explode, owner)
			return true
		else
			-- bounce bombs away without any damage
			if (explode.type == MT_PLAGUE_BOMB)
				local touchspeed = P_AproxDistance(explode.momx, explode.momy)
				if (touchspeed < mo.scale)
					touchspeed = mo.scale
				end

				P_InstaThrust(explode, mo.angle, touchspeed)
				--explode.momz = -$1

				S_StartSound(explode, mo.info.painsound)
			end

			-- Whenever it's a bomb or an explosion, back up because it's dangerous!
			if (mo.target and mo.target.valid)
				mo.target.extravalue1 = -mo.target.info.speed
			end

			return false
		end
	elseif ((mo.type == MT_EGGMOBILE4) and (mo.movefactor == 0 and mo.extravalue1 == 0))
		-- Sequence breaking messes this boss up real bad!
		-- The ground around him is broken based on his current health,
		-- which means weird chunks go away when the cage slams down.
		-- Besides, hitting him through the cage is kind of pants.

		-- To explain, since this boss is barely documented:
		-- movefactor is the boss's cage Z offset -- if this is 0, then the cage is down.
		-- extravalue1 is set when the boss has no cage -- if this is 0, then there is a cage.
		return false
	elseif (mo.type == MT_SPIKE) or (mo.type == MT_WALLSPIKE)
		--[[
		if (mo.health > 0)
			P_KillMobj(mo, explode, explode.target)
		end
		--]]

		-- nah, that's unsatisfying, just explode with no effect on spikes
		return true
	else
		local battleBashed = false

		if (CBW_Battle)
			if (mo.battleobject) and not (mo.sentient)
				-- Push away battle mode bashables
				explode.battle_atk = 3
				explode.battle_def = 1

				explode.pushed = mo
				mo.pushed = explode

				CBW_Battle.DoPlayerInteract(mo, explode)
				battleBashed = true
			end
		end

		-- Basic flag fallbacks
		if (mo.flags & MF_SHOOTABLE)
			-- detonate bomb, cause damage
			P_DamageMobj(mo, explode, owner, 1)
			return true
		elseif (mo.flags & MF_PUSHABLE)
		or ((mobjinfo[mo.type].flags & MF_PUSHABLE) and mo.fuse) -- allows pushing rollout rocks while they're being ridden
			-- detonate bomb, push around
			local scaling = FixedDiv(explode.scale, mo.scale)
			local thrust = FixedMul(mo.info.speed, scaling)
			local minthrust = 4*scaling

			if (thrust < minthrust)
				thrust = minthrust
			end

			local hang = R_PointToAngle2(
				explode.x - explode.momx,
				explode.y - explode.momy,
				mo.x - mo.momx,
				mo.y - mo.momy
			)
			local vang = R_PointToAngle2(
				0, 0,
				P_AproxDistance(
					(explode.x - explode.momx) - (mo.x - mo.momx),
					(explode.y - explode.momy) - (mo.y - mo.momy)
				),
				(mo.z - mo.momz) - (explode.z - explode.momz)
			)

			mo.momx = FixedMul(FixedMul(thrust, cos(hang)), cos(vang)) -- + $1
			mo.momy = FixedMul(FixedMul(thrust, sin(hang)), cos(vang)) -- + $1
			mo.momz = FixedMul(thrust, sin(vang)) -- + $1

			S_StartSound(mo, mo.info.activesound)
			mo.target = explode -- might not need this?
			return true
		elseif (mo.flags & MF_SPRING)
		and (explode.type == MT_PLAGUE_BOMB)
		and (battleBashed == false)
			-- bombs take springs!
			P_DoSpring(mo, explode)
			return false
		elseif (mo.flags & (MF_ENEMY|MF_BOSS|MF_SOLID))
		or (mo.type == MT_FAKEMOBILE)
		or (mo.type == MT_EGGROBO1)
			if (battleBashed == false)
				-- bounce bombs away without any damage
				if (explode.type == MT_PLAGUE_BOMB)
					local hitTop = false
					local topRange = mo.height / 4
					local flip = P_MobjFlip(explode)

					if (flip == -1)
						if ((explode.z + explode.height - explode.momz) < mo.z + topRange)
							hitTop = true
						end
					else
						if ((explode.z - explode.momz) > (mo.z + mo.height) - topRange)
							hitTop = true
						end
					end

					if (hitTop == true)
						-- Offset by one, to prevent c stack overflow
						if (flip == -1)
							explode.z = mo.z - explode.height - 1
						else
							explode.z = mo.z + mo.height + 1
						end

						Plague.ApplyMomZ(explode, 4*FRACUNIT)
					else
						explode.angle = R_PointToAngle2(mo.x, mo.y, explode.x, explode.y)
					end

					P_InstaThrust(explode, explode.angle, FixedMul(explode.info.speed, explode.scale))
				end

				if (mo.type == MT_METALSONIC_BATTLE) and (mo.state == mo.info.spawnstate)
					-- vectorizes metal sonic without detonating the bomb
					mo.flags2 = $1|MF2_CLASSICPUSH
				end
			end
		end

		return battleBashed
	end

	return false
end

Plague.ExplodeSearch = function(explode, skipobjs)
	if (explode.health <= 0)
		return false
	end

	local damaged = false
	local rad = explode.radius

	if (skipobjs != true)
		searchBlockmap("objects", function(ref, found)
			if (ref.target == found) then return end

			if (found.z + found.height < ref.z)
			or (found.z > ref.z + ref.height)
				return
			end

			local xdist = abs(found.x - ref.x)
			local ydist = abs(found.y - ref.y)

			if (xdist > found.radius + rad)
			or (ydist > found.radius + rad)
				return
			end

			damaged = Plague.ExplosionDamage(ref, found)
		end, explode)
	end

	if (Plague.BustFOFsInSector(explode, explode.subsector.sector)) then
		damaged = true
	end

	local minx = explode.x
	local maxx = explode.x
	local miny = explode.y
	local maxy = explode.y

	-- Also subtracts because of Dumb MF_BOUNCE Shit (please add wall collision hooks)
	if (explode.momx)
		if (explode.x - explode.momx > explode.x + explode.momx)
			minx = explode.x + explode.momx
			maxx = explode.x - explode.momx
		else
			minx = explode.x - explode.momx
			maxx = explode.x + explode.momx
		end
	end

	if (explode.momy)
		if (explode.y - explode.momy > explode.y + explode.momy)
			miny = explode.y + explode.momy
			maxy = explode.y - explode.momy
		else
			miny = explode.y - explode.momy
			maxy = explode.y + explode.momy
		end
	end

	local extra = rad + (64*FRACUNIT)

	minx = $1 - extra
	maxx = $1 + extra
	miny = $1 - extra
	maxy = $1 + extra

	searchBlockmap("lines", function(ref, found)
		local pointx, pointy = P_ClosestPointOnLine(ref.x, ref.y, found)

		if (ref.momx or ref.momy)
			local success = false

			local testx = ref.x
			local testy = ref.y

			local xdist = abs(pointx - testx)
			local ydist = abs(pointy - testy)

			local dist = P_AproxDistance(xdist, ydist)

			if (xdist <= rad)
			and (ydist <= rad)
				success = true
			else
				local dir = R_PointToAngle2(0, 0, ref.momx, ref.momy)

				local c = FixedMul(rad, cos(dir))
				local s = FixedMul(rad, sin(dir))

				if (c < cos(dir)) then c = cos(dir) end
				if (s < sin(dir)) then s = sin(dir) end

				while (success == false)
					testx = testx + c
					testy = testy + s

					local newxdist = abs(pointx - testx)
					local newydist = abs(pointy - testy)
					local newdist = P_AproxDistance(newxdist, newydist)

					if (newdist >= dist)
						break
					end

					xdist = newxdist
					ydist = newydist
					dist = newdist

					if (xdist <= rad)
					and (ydist <= rad)
						success = true
						break
					end
				end
			end

			if (success == false)
				return
			end
		else
			local xdist = abs(pointx - ref.x)
			local ydist = abs(pointy - ref.y)

			if (xdist > rad)
			or (ydist > rad)
				return
			end
		end

		--if (found.backsector and found.backsector.valid)
		--and not (P_PointOnLineSide(ref.x, ref.y, found))
			if (Plague.BustFOFsInSector(ref, found.backsector)) then damaged = true end
		--else
			if (Plague.BustFOFsInSector(ref, found.frontsector)) then damaged = true end
		--end
	end, explode, minx, maxx, miny, maxy)

	return damaged
end

Plague.SpawnPlagueExplosion = function(mo, pmo, x, y, z, delay, nosound)
	local explode = P_SpawnMobjFromMobj(mo, x, y, z, MT_PLAGUE_EXPLOSION)
	explode.target = pmo

	explode.z = $1 - (explode.height / 2)

	if (pmo and pmo.valid) and (pmo.player and pmo.player.valid)
		explode.color = pmo.player.skincolor
	else
		explode.color = SKINCOLOR_SEAFOAM
	end

	Plague.ExplodeSearch(explode, false)

	if (delay)
		explode.state = S_PLAGUE_EXPLOSIONDELAYED
		explode.tics = delay
	end

	if not (nosound)
		S_StartSound(explode, sfx_pkexpl)
	end
end

addHook("MobjThinker", function(mo)
	if (mo.state == S_PLAGUE_EXPLOSION1)
	or (mo.state == S_PLAGUE_EXPLOSION2)
	or (mo.state == S_PLAGUE_EXPLOSION3)
		Plague.ExplodeSearch(mo, false)
	end
end, MT_PLAGUE_EXPLOSION)
