--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- main.lua: Main gameplay loop
--

Plague.Main = function(player)
	local onground = P_IsObjectOnGround(player.mo)

	if (player.mo.plague == nil)
		player.mo.plague = {
			buttonsheld = (player.pflags & (PF_JUMPDOWN|PF_SPINDOWN)),
			spinbt = 0,

			bombburst = false,
			burstcharge = 0,
			burstinvuln = 0,
			burstflash = false,

			bombs = {},
			bombthrowcool = 0,
			bombthrowing = 0,
			throwangle = player.mo.angle,

			djumped = false,
			paindelay = Plague.BURSTINVULTIME,
			skidtime = 0,

			prevonground = onground,
			prevcarry = player.powers[pw_carry],

			danceheld = 0,
			danceangle = player.mo.angle,

			nodeathanim = false,
			deathanim = 0,
			deathz = 0,

			surgeangle = player.mo.angle,
		}

		-- Ensure this is reset on map load
		player.mo.color = player.skincolor
		player.mo.colorized = false
	end

	-- reset death vars
	player.mo.plague.nodeathanim = false
	player.mo.plague.deathanim = 0
	player.mo.plague.deathz = 0

	if (player.powers[pw_carry] == CR_NIGHTSMODE)
		player.mo.plague.bombburst = false
		player.mo.plague.burstcharge = 0
		player.mo.plague.burstinvuln = 0
		player.mo.plague.bombthrowcool = 0
		player.mo.plague.bombthrowing = 0
		player.mo.plague.throwangle = player.mo.angle

		player.mo.plague.djumped = true
		player.mo.plague.paindelay = Plague.BURSTINVULTIME
		player.mo.plague.skidtime = 0

		player.mo.plague.prevonground = onground
		player.mo.plague.prevcarry = CR_NIGHTSMODE

		player.mo.plague.danceheld = 0
		player.mo.plague.danceangle = player.mo.angle

		player.mo.plague.surgeangle = player.mo.angle

		if (player.mo.plague.burstflash)
			player.mo.color = player.skincolor
			player.mo.colorized = false
			player.mo.plague.burstflash = false
		end

		return
	end

	if (player.charflags & SF_DASHMODE)
		-- This allows dash mode if it's forced onto him (tether, neo sonic)
		player.runspeed = (skins[player.mo.skin].normalspeed * 7) / 9
	else
		-- Otherwise runspeed is supposed to be stupid high
		-- because the running animation playback speed looks dumb.
		player.runspeed = skins[player.mo.skin].runspeed
	end

	-- Handle slippery luigi physics
	Plague.Friction(player, onground)
	Plague.Skid(player, onground)

	-- Reset flags
	if (onground) or (player.powers[pw_carry])
		player.mo.plague.bombburst = false
		player.mo.plague.djumped = false
	elseif not (player.mo.state == S_PLAY_PLAGUEBURST or player.panim == PA_FALL)
		player.mo.plague.bombburst = false
	end

	-- Bomb burst handling
	if (player.mo.plague.burstinvuln > 0) then player.mo.plague.burstinvuln = $1-1 end
	Plague.BombBurst(player)

	-- Bomb throw handling
	if (player.mo.plague.bombthrowcool > 0) then player.mo.plague.bombthrowcool = $1-1 end

	if (P_PlayerInPain(player))
		-- Only interruption to bomb throwing
		player.mo.plague.bombthrowing = 0
	elseif (player.mo.plague.bombthrowing > 0)
		-- few tics to ensure nothing stupid like falling states immediately overwriting this state
		local st = S_PLAY_PLAGUETHROWAIR
		if (onground)
			st = S_PLAY_PLAGUETHROWGROUND
		end

		if (player.mo.state != st)
			player.mo.state = st
		end

		if not ((player.mo.flags2 & MF2_TWOD) or (twodlevel) or (player.mo.currentaxis))
			player.mo.plague.throwangle = player.mo.angle
		end

		player.mo.plague.bombthrowing = $1-1
		if (player.mo.plague.bombthrowing <= 0)
			Plague.SpawnBomb(player, onground)
		end
	end

	if (player.mo.plague.spinbt)
	and not (player.mo.plague.buttonsheld & PF_SPINDOWN)
	and (player.mo.plague.bombthrowcool <= 0)
	and not ((player.pflags & PF_SPINNING) and P_IsObjectOnGround(player.mo) == true) -- bomb throwing only interrupts spin state when you're in the air
	and not (P_SuperReady(player) and not onground)
		Plague.ThrowBomb(player, onground)
	end

	if not (player.pflags & PF_THOKKED)
		-- Restore djump if another script restores ability :)
		player.mo.plague.djumped = false;
	end

	-- Double jump
	if (Plague.AllowAbilities(player, PF_JUMPSTASIS, true))
	and (Plague.DoesntHaveFlag(player) == true)
		if (player.cmd.buttons & BT_JUMP)
		and not (player.mo.plague.buttonsheld & PF_JUMPDOWN)
		and (player.mo.plague.djumped == false)
		and not (onground or player.mo.plague.prevonground)
			player.pflags = $1 & ~PF_JUMPED

			P_DoJump(player, false)
			player.mo.state = S_PLAY_PLAGUEDOUBLEJUMP

			S_StartSound(player.mo, sfx_jump) -- Two sounds is not a mistake
			S_StartSound(player.mo, sfx_pkdjmp)

			player.mo.plague.djumped = true
			player.pflags = $1 | (PF_STARTJUMP|PF_THOKKED)
		end
	end

	-- Painstate
	if (P_PlayerInPain(player))
		player.mo.plague.bombburst = false
		player.mo.plague.bombthrowing = 0

		-- EXTREME knockback
		if (player.mo.plague.paindelay >= Plague.BURSTINVULTIME)
			player.mo.momx = $1*3
			player.mo.momy = $1*3
			player.mo.plague.paindelay = Plague.BURSTINVULTIME-1
		elseif (player.mo.plague.paindelay > 0)
			player.mo.plague.paindelay = $1-1
		end
	else
		player.mo.plague.paindelay = Plague.BURSTINVULTIME
	end

	-- Reduced control for initial bomb burst
	if (player.mo.state == S_PLAY_PLAGUEBURST)
		player.mo.movefactor = $1 / 3
		if (player.mo.momx or player.mo.momy)
			player.drawangle = R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy)
		else
			player.drawangle = player.mo.angle
		end
	end

	-- Extra gravity while bomb bursting
	if (player.mo.plague.bombburst)
		player.mo.momz = $1 + (P_GetMobjGravity(player.mo) / 2)
	end

	-- Go into fall state
	if (player.mo.state == S_PLAY_PLAGUEBURST)
	and (player.mo.momz * P_MobjFlip(player.mo) < 0)
		player.mo.state = S_PLAY_FALL
	end

	-- Come to a natural stop while throwing
	if (player.mo.state == S_PLAY_PLAGUETHROWGROUND)
	or (player.mo.state == S_PLAY_PLAGUETHROWAIR)
		player.mo.movefactor = 0
		player.drawangle = player.mo.plague.throwangle
	end

	if (player.mo.state == S_PLAY_WAIT)
		player.mo.state = S_PLAY_PLAGUEIDLE1
	end

	-- "Crouch" by holding Custom 1
	if (Plague.CanDance(player) == true) and (player.cmd.buttons & BT_CUSTOM1)
		if (player.mo.plague.danceheld > (TICRATE/3))
			local numDanceFrames = 10
			local numDanceFramesPerLeg = numDanceFrames / 2

			if (Plague.InDanceState(player) == false)
				player.mo.state = S_PLAY_PLAGUEDANCESTART
			end

			player.drawangle = player.mo.plague.danceangle

			player.mo.momx = 0
			player.mo.momy = 0

			if (player.mo.state == S_PLAY_PLAGUEDANCELOOP)
				local danceFrame = (player.mo.frame & FF_FRAMEMASK) % numDanceFramesPerLeg

				if (danceFrame == 2 or danceFrame == 3)
					local numTics = 6
					local rotateAmount = ANGLE_90 / numTics
					player.mo.plague.danceangle = $1 + rotateAmount

					if (Plague.AllowAbilities(player, PF_STASIS, false) == true)
						P_InstaThrust(player.mo, player.mo.angle, 10 * player.mo.scale)
					end
				end
			end
		else
			player.mo.state = S_PLAY_PLAGUECROUCH
			player.mo.plague.danceangle = player.drawangle
			player.mo.plague.danceheld = $1+1
		end
	else
		player.mo.plague.danceheld = 0

		if (Plague.InDanceState(player) == true)
			player.mo.state = S_PLAY_STND
		end
	end

	-- Bomb Burst color flashing
	if (player.mo.plague.burstcharge >= Plague.BURSTTHRESHOLD) and not (leveltime & 1)
		player.mo.color = Plague.OppositeColor(player)
		player.mo.colorized = true
		player.mo.plague.burstflash = true
	elseif (player.mo.plague.burstflash)
		if (player.powers[pw_shield] & SH_FIREFLOWER)
			player.mo.color = SKINCOLOR_WHITE
		else
			player.mo.color = player.skincolor
		end
		player.mo.colorized = false
		player.mo.plague.burstflash = false
	end

	-- Landing sound
	if (onground) and not (player.mo.plague.prevonground)
		S_StartSound(player.mo, sfx_pkland)
	end

	-- Manually do this since it doesn't work for me for some reason...
	player.mo.plague.buttonsheld = (player.pflags & (PF_JUMPDOWN|PF_SPINDOWN));

	player.mo.plague.prevonground = onground
	player.mo.plague.prevcarry = player.powers[pw_carry]
end

Plague.Death = function(player)
	if (player.mo.plague.burstflash)
		player.mo.color = player.skincolor
		player.mo.colorized = false
		player.mo.plague.burstflash = false
	end

	if (player.mo.plague.nodeathanim) -- don't do this for drowning
		return
	end

	--player.mo.color = skins[player.mo.skin].supercolor + (player.mo.plague.deathanim % 4)
	--player.mo.frame = $1|FF_FULLBRIGHT

	player.mo.momx = 0
	player.mo.momy = 0

	if (player.mo.plague.deathanim == Plague.DEATHTIME)
		local explode = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_PLAGUE_DEATHEXPLODE)
		explode.color = player.skincolor
		player.mo.plague.deathanim = Plague.DEATHTIME+1
		player.mo.plague.deathz = player.mo.z
	elseif (player.mo.plague.deathanim > Plague.DEATHTIME)
		player.mo.flags = ($1 & ~MF_NOCLIPHEIGHT)|MF_NOGRAVITY
		player.mo.flags2 = $1|MF2_DONTDRAW
		player.mo.momz = 0
		player.mo.z = player.mo.plague.deathz
	else
		player.mo.plague.deathanim = $1+1
		player.mo.momz = $1 + (P_GetMobjGravity(player.mo) / 2)
		if (player.mo.momz * P_MobjFlip(player.mo) <= 0)
			player.mo.flags = $1 & ~MF_NOCLIPHEIGHT
		end
	end
end

Plague.UpdatePAnim = function(player)
	if (player.mo.state == S_PLAY_PLAGUEBURST)
		player.panim = PA_ABILITY
	elseif (player.mo.state == S_PLAY_PLAGUETHROWGROUND)
	or (player.mo.state == S_PLAY_PLAGUETHROWAIR)
		player.panim = PA_ABILITY2
	elseif (player.mo.state == S_PLAY_PLAGUEDOUBLEJUMP)
		player.panim = PA_JUMP -- PA_SPRING?
	elseif (player.mo.state == S_PLAY_PLAGUECROUCH)
	or (player.mo.state >= S_PLAY_PLAGUEIDLE1 and player.mo.state <= S_PLAY_PLAGUEIDLE6)
	or (Plague.InDanceState(player))
		player.panim = PA_IDLE
	end
end

addHook("MobjThinker", function(mo)
	if not (mo.player and mo.player.valid)
		return
	end

	local player = mo.player

	-- ensure it's set *before* running thinker
	Plague.UpdatePAnim(player)

	if (mo.skin == "plagueknight")
		if (player.playerstate == PST_DEAD)
			Plague.Death(player)
		elseif (player.playerstate == PST_LIVE) and (mo.health)
			Plague.Main(player)
		end
	elseif (mo.plague != nil)
		mo.plague = nil
	end

	-- AND again, if other Lua/hardcode checks it, if we updated state in plagueKnightThinker
	Plague.UpdatePAnim(player)

	if (mo.coopburst != nil)
		if (P_IsObjectOnGround(mo))
			mo.coopburst.time = 0
		end

		if (mo.coopburst.time > 0)
			mo.coopburst.time = $1 - 1

			local radius = (mo.radius / mo.scale) * 2
			local aura = P_SpawnMobjFromMobj(
				mo,
				P_RandomRange(-radius, radius) * FRACUNIT,
				P_RandomRange(-radius, radius) * FRACUNIT,
				(P_RandomRange(-radius, radius) * FRACUNIT) + FixedDiv(mo.height / 2, mo.scale),
				MT_PLAGUE_SURGEAURA
			)
			aura.color = mo.coopburst.color
		end
	end
end, MT_PLAYER)

addHook("PlayerThink", function(player)
	-- This shit again?
	-- This time so that the idle/crouch/dance animations don't break you out of Amy's grasp.
	if (player.mo and player.mo.valid and player.mo.skin == "plagueknight")
		Plague.UpdatePAnim(player)
	end
end)

addHook("PreThinkFrame", function()
	for player in players.iterate do
		if not (player.mo and player.mo.valid)
			continue
		end

		if (player.mo.skin != "plagueknight")
			continue
		end

		if (player.mo.plague == nil)
			continue
		end

		if (player.powers[pw_carry] == CR_ROPEHANG)
			if (player.cmd.buttons & BT_SPIN)
				-- Allows you to charge bomb burst while hanging off ropes
				player.mo.plague.spinbt = BT_SPIN
				player.cmd.buttons = $1 & ~BT_SPIN
			else
				player.mo.plague.spinbt = 0
			end
		else
			player.mo.plague.spinbt = (player.cmd.buttons & BT_SPIN)
		end
	end
end)

addHook("MobjDeath", function(mo, inf, src, dt)
	if (mo.plague != nil)
		if (dt == DMG_DROWNED or dt == DMG_SPACEDROWN)
			mo.plague.nodeathanim = true
		end

		-- Set draw angle since pain has angles
		-- Yes, this is R_PointToAngle, the original.
		-- Should HOPEFULLY not cause netplay desyncs because this is on death,
		-- unless if another script is being really stupid with it.
		mo.player.drawangle = R_PointToAngle(mo.x, mo.y) + ANGLE_180
	end
end, MT_PLAYER)

addHook("ShouldDamage", function(mo, inf, src, dmg, dt)
	if (mo.plague != nil) and (mo.plague.burstinvuln > 0) and not (dt & DMG_DEATHMASK)
		return false
	end
end, MT_PLAYER)

addHook("MobjDamage", function(mo, inf, src, dmg, dt)
	if (inf and inf.valid)
	and (inf.type == MT_PLAGUE_BOMB or inf.type == MT_PLAGUE_EXPLOSION)
		-- Plague Knight's bombs do simple knockback in Race/Ringslinger.
		-- They only deal actual damage in Battle.

		if (CBW_Battle)
			if (src and src.valid)
				CBW_Battle.PlayerCreditPusher(mo.player, src)
			end
		else
			P_DoPlayerPain(mo.player, explode, owner)
			mo.player.powers[pw_flashing] = 0 -- Allows juggling
			return true
		end
	end
end, MT_PLAYER)
