--
-- ==== Plague Knight ====
-- == By TehRealSalt ==
--
-- slip.lua: Slippery physics modifiers
--

Plague.AutoBrake = function(player)
	-- Double-apply autobrake, since autobrake doesn't really work for some reason with Plague?
	if (player.powers[pw_noautobrake])
		return;
	end

	if (player.powers[pw_carry] or player.powers[pw_nocontrol])
		return;
	end

	if ((player.pflags & (PF_AUTOBRAKE|PF_APPLYAUTOBRAKE|PF_STASIS)) != (PF_AUTOBRAKE|PF_APPLYAUTOBRAKE))
		return;
	end

	if (player.cmd.forwardmove or player.cmd.sidemove)
		return;
	end

	if not (player.rmomx or player.rmomy)
		return;
	end

	if (player.capsule and player.capsule.valid)
	and (player.capsule.reactiontime == #player + 1)
		return;
	end

	local acceleration = (player.accelstart + (FixedDiv(player.speed, player.mo.scale)>>FRACBITS) * player.acceleration) * player.thrustfactor * 20;
	local moveAngle = R_PointToAngle2(0, 0, player.rmomx, player.rmomy);

	acceleration = FixedMul(acceleration<<FRACBITS, player.mo.movefactor)>>FRACBITS;

	P_Thrust(player.mo, moveAngle, -acceleration);
end

Plague.Friction = function(player, onground)
	player.mo.friction = $1 + 3072

	if (player.mo.friction > FRACUNIT) then player.mo.friction = FRACUNIT end
	if (player.mo.friction < 0) then player.mo.friction = 0 end

	if (onground) and (player.mo.friction > 0)
		player.mo.movefactor = FixedDiv(ORIG_FRICTION, player.mo.friction)

		if (player.mo.movefactor < FRACUNIT)
			player.mo.movefactor = (8*$1) - (7*FRACUNIT)
		else
			player.mo.movefactor = FRACUNIT
		end

		Plague.AutoBrake(player);
	end
end

Plague.Skid = function(player, onground)
	-- Does soundless, particle-less, slippery skidding.
	if not (onground)
	or (player.pflags & (PF_JUMPED|PF_SPINNING|PF_SLIDING))
	or (player.mo.eflags & MFE_GOOWATER)
	or (player.mo.state == S_PLAY_PLAGUETHROWGROUND)
	or (player.mo.state == S_PLAY_PLAGUETHROWAIR)
		player.mo.plague.skidtime = 0
		player.skidtime = 0
	elseif (player.mo.plague.skidtime)
		if (player.mo.state != S_PLAY_SKID)
			player.mo.state = S_PLAY_SKID
			player.mo.tics = player.mo.plague.skidtime
		end

		player.skidtime = player.mo.plague.skidtime
		player.mo.plague.skidtime = $1 - 1
	else
		local pmx = player.rmomx + player.cmomx
		local pmy = player.rmomy + player.cmomy

		if (P_AproxDistance(pmx, pmy) >= 10*player.mo.scale)
		and (player.mo.momx != pmx or player.mo.momy != pmy)
		and (P_GetPlayerControlDirection(player) == 2)
			local mang = R_PointToAngle2(0, 0, pmx, pmy)
			local pang = R_PointToAngle2(pmx, pmy, player.mo.momx, player.mo.momy)
			local dang = abs(mang - pang)

			if (dang > ANGLE_157h)
				if (player.mo.state != S_PLAY_SKID)
					player.mo.state = S_PLAY_SKID
				end

				player.mo.plague.skidtime = FixedDiv(35 << (FRACBITS-1), FixedSqrt(player.mo.movefactor)) >> FRACBITS
				player.mo.tics = player.mo.plague.skidtime
				player.skidtime = player.mo.plague.skidtime
			end
		end
	end
end
